#include "RRTExpansion.hpp"

#include "API/Device/robot.hpp"
#include "API/ConfigSpace/configuration.hpp"
#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"
#include "API/Grids/PointCloud.hpp"

#include "P3d-pkg.h"
#include "Planner-pkg.h"
#include "Collision-pkg.h"

using namespace std;
using namespace tr1;


/**
 * Sample a direction for the expansion process (taking into account the goal bias, if relevant).
 * @param directionNode: contains the goal node when relevant
 * @return the configuration toward which the tree will be expanded
 */
confPtr_t RRTExpansion::sampleExpansionDirection(Node * directionNode)
{
    // bias toward the goal configuration
    if (ENV.getBool(Env::isGoalBiased) && p3d_random(0., 1.) <= ENV.getDouble(Env::Bias))
    {
        directionNode = goalNode;
        return goalNode->getConfiguration();
    }
    // or random sampling in the entire C-Space
    else
        return m_Graph->getRobot()->shoot();
}


/**
 * RRT Extend.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned RRTExpansion::extend(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{
	unsigned nbCreatedNodes(0);

	// perform an extension toward directionConf
	shared_ptr<LocalPath> directionPath = shared_ptr<LocalPath> (
			new LocalPath(fromNode->getConfiguration(), directionConf));
	double positionAlongDirection(0.);
	shared_ptr<LocalPath> extensionLocalpath;
	bool extensionSucceeded = nextStep(*directionPath, directionNode, positionAlongDirection,
			extensionLocalpath, Env::Extend);

	// refinement control
	if (extensionSucceeded && ENV.getBool(Env::refinementControl) &&
			!this->refinementControl(*directionPath, positionAlongDirection, *fromNode))
		return 0;

	// add node to graph, if everything succeeded
	if (extensionSucceeded)
		m_last_added_node = addNode(fromNode, *extensionLocalpath, positionAlongDirection, directionNode,
				nbCreatedNodes);
	else
		expansionFailed(*fromNode);

	return nbCreatedNodes;
}


/**
 * RRT Connect.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned RRTExpansion::connect(Node * expansionNode, confPtr_t directionConfig, Node * directionNode)
{
	bool extensionSucceeded(false);
	bool failed(false);
	unsigned nbCreatedNodes(0);
	Node* fromNode = expansionNode;
	Node* extensionNode(NULL);
	shared_ptr<LocalPath> directionLocalpath;
	double positionAlongDirection(0.);
	shared_ptr<LocalPath> extensionLocalpath;

	// Perform extension toward directionConfig
	// Additional nodes creation in the nExtend case, but without checking for expansion control
	while (!failed && positionAlongDirection < 1.)
	{
		directionLocalpath = shared_ptr<LocalPath> (
				new LocalPath(fromNode->getConfiguration(), directionConfig));

		// Expand one step along the local path "extensionLocalpath"
		extensionSucceeded = nextStep(*directionLocalpath, directionNode, positionAlongDirection,
				extensionLocalpath, Env::Connect);

		failed |= !extensionSucceeded;

		// Expansion Control
		if (!failed) {
			if (ENV.getBool(Env::refinementControl) && !this->refinementControl(*directionLocalpath,
					positionAlongDirection, *expansionNode)) {
				failed = true;
			}
		}
		// Add node to graph if everything succeeded
		if (!failed) {
			extensionNode = addNode(fromNode, *extensionLocalpath, positionAlongDirection, directionNode,
					nbCreatedNodes);
			m_last_added_node = extensionNode;
		}
		if (failed) {
			expansionFailed(*expansionNode);
		}

		if (!failed) {
			fromNode = extensionNode;
		}
	}

	return nbCreatedNodes;
}


/**
 * RRT Expand process: call the appropriate expansion method.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @param method: Extend or Connect
 * @return the number of created nodes
 */
unsigned RRTExpansion::expandProcess(Node * fromNode, confPtr_t directionConf, Node * directionNode,
        Env::expansionMethod method)
{
    switch (method)
    {
    case Env::Extend:
        return this->extend(fromNode, directionConf, directionNode);

    case Env::Connect:
        return this->connect(fromNode, directionConf, directionNode);

    default:
        cerr << "*** ERROR: RRT-Expansion::expandProcess --> method not implemented !!! ***" << endl;
        return 0;
    }
}


/**
 * Attempt to link two given nodes (i.e. two components, when relevant).
 * @return TRUE if the junction is successful.
 */
bool RRTExpansion::attemptJunction(Node * fromNode, Node * toNode)
{
    confPtr_t fromConf = fromNode->getConfiguration();
    confPtr_t toConf = toNode->getConfiguration();

    // path along which the junction is attempted
    LocalPath path(fromConf, toConf);
    double pathLength = path.getParamMax();

    // N.B.: the connection is attempted only if the path length is less than a specified threshold
    // (which is some multiple of the extension step)
    if (pathLength <= ENV.getDouble(Env::minimalFinalExpansionGap) * this->step())
    {
        // check whether the path is valid (i.e. collision-free)
        return path.isValid();
    }

    return false;
}


shared_ptr<Configuration> RRTExpansion::getExpansionDirection(Node* expandComp, Node* goalComp,
        bool samplePassive, Node*& directionNode)
{
    if (p3d_GetCostMethodChoice() == MONTE_CARLO_SEARCH)
    {
        confPtr_t q = m_Graph->getRobot()->shootDir(samplePassive);

        p3d_addConfig(m_Graph->getRobot()->getRobotStruct(),
                      q->getConfigStruct(),
                      expandComp->getConnectedComponent()->getCompcoStruct()->dist_nodes->N->q,
                      q->getConfigStruct());

        return q;
    }

    shared_ptr<Configuration> q;
    int savedRlg;

    if (m_IsDirSampleWithRlg)
    {
        // Save the previous Rlg setting to shoot without Rlg
        savedRlg = p3d_get_RLG();
        p3d_set_RLG(false);
    }

    // Selection in the entire CSpace and
    // biased to the Comp of the goal configuration
    if (ENV.getBool(Env::isGoalBiased) && p3d_random(0., 1.) <= ENV.getDouble(Env::Bias))
    {
        // select randomly a node in the goal component as direction of expansion
        directionNode = m_Graph->randomNodeFromComp(goalComp);
        q = directionNode->getConfiguration();
    }
    else
    {
        switch (m_ExpansionDirectionMethod)
        {
            case SUBREGION_CS_EXP:
                // Selection in a subregion of the CSpace
                // (typically close to the current tree)
                // and  biased to the goal configuration
                q = shared_ptr<Configuration> (
                        new Configuration(m_Graph->getRobot()));
						
#ifdef P3D_PLANNER
                p3d_shoot_inside_box(m_Graph->getRobot()->getRobotStruct(),
                                     /*expandComp->getConfiguration()->getConfigStruct(),*/
                                     q->getConfigStruct(), 
																		 expandComp->getConnectedComponent()->getCompcoStruct()->box_env_small,
                                     (int) samplePassive);
#else
						printf("P3D_PLANNER not compiled in %s in %s",__func__,__FILE__);
#endif
						
            break;
#ifdef LIGHT_PLANNER
			case NAVIGATION_BEFORE_MANIPULATION:
			{
				m_Graph->getRobot()->deactivateCcConstraint();
				
				q = m_Graph->getRobot()->shoot();
				Node* closest = m_Graph->nearestWeightNeighbour(expandComp,q,false,ONLY_ROBOT_BASE);
				
				LocalPath path(closest->getConfiguration(),q);
//				cout << "Param max = " << path.getParamMax() << " , step = " << step() << endl;
				q = path.configAtParam(2*step());
				
//				cout << "1 :" << endl;
//				q->print();
				
				m_Graph->getRobot()->setAndUpdate(*q);
				q = m_Graph->getRobot()->shootAllExceptBase();
				q->setConstraintsWithSideEffect();
				
//				cout << "2 :" << endl;
//				q->print();
				
//				m_Graph->getRobot()->setAndUpdate(*q);
//				g3d_draw_allwin_active();
				
				m_Graph->getRobot()->activateCcConstraint();
//				q->setConstraintsWithSideEffect();
//				m_Graph->getRobot()->setAndUpdate(*q);
//				g3d_draw_allwin_active();
			}
			break;
#endif
			case GLOBAL_CS_EXP:
          default:
            // Selection in the entire CSpace
            q = m_Graph->getRobot()->shoot(samplePassive);
        }
    }
    if (!m_IsDirSampleWithRlg)
    {
        //Restore the previous Rlg setting
        p3d_set_RLG(savedRlg);
    }
    return (q);
}

Node* RRTExpansion::getExpansionNode(Node* compNode, confPtr_t direction, int distance)
{
    if (p3d_GetCostMethodChoice() == MONTE_CARLO_SEARCH)
    {
        return m_Graph->getNode(compNode->getConnectedComponent()->getCompcoStruct()->dist_nodes->N);
    }

    int K;
    int KNearest = -1;
    int NearestPercent;
    vector<Node*> KNearestNeighbors;

    switch (distance)
    {

    case NEAREST_EXP_NODE_METH:
        // choose the nearest node of the connected component
        return (m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance));

    case K_NEAREST_EXP_NODE_METH:
        // randomly select one of the K nearest nodes of the connected component
        K = ceil(compNode->getConnectedComponent()->getNumberOfNodes() / 100.);
        if (K == 1)
            return m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance);
        else {
            KNearestNeighbors = compNode->getConnectedComponent()->KNearestWeightNeighbour(direction, K,
                    numeric_limits<double>::max(), p3d_GetIsWeightedChoice(), distance);
            unsigned index = (int) floor(p3d_random(0.0, (double) KNearestNeighbors.size() - EPS6));
            return KNearestNeighbors[index];
        }

    case BEST_SCORE_EXP_METH:
        /* Select the node which has the best score: weight*dist */
        return (m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance));

    case K_BEST_SCORE_EXP_METH:
        NearestPercent = m_kNearestPercent;
        KNearest
                = MAX(1,(int)((NearestPercent*(compNode->getConnectedComponent()->getNumberOfNodes()))/100.));
        return (m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance));

    case RANDOM_IN_SHELL_METH:
        /* Select randomly among all the nodes inside a given portion of shell */
        return (m_Graph->getNode(
                hrm_selected_pb_node(m_Graph->getGraphStruct(), direction->getConfigStruct(),
                        compNode->getNodeStruct()->comp)));

    case RANDOM_NODE_METH:
        return (m_Graph->getNode(p3d_RandomNodeFromComp(compNode->getConnectedComponent()->getCompcoStruct())));

    default:
        /* By default return the nearest node of the component */
        return (m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance));
    }
}
