#ifndef RRTEXPANSION_HPP
#define RRTEXPANSION_HPP

#include "planner/Diffusion/Variants/BaseExpansion.hpp"

/**
  @ingroup Diffusion
  * RRT Expansion Methods
  *
  * Functions to expand a tree in the CSpace
  * Collision Checking, Random sampling and Biasing
  */
class RRTExpansion: public BaseExpansion
{

public:
    RRTExpansion() : BaseExpansion() {};
    RRTExpansion(Graph * G) : BaseExpansion(G) {};

	//! Constructor.
    RRTExpansion(Graph * G, Node * goal) : BaseExpansion(G), goalNode(goal) {};

    //! Destructor.
    virtual ~RRTExpansion() {};

    /**
     * Sample a direction for the expansion process (taking into account the goal bias, if relevant).
     * @param directionNode: contains the goal node when relevant
     * @return the configuration toward which the tree will be expanded
     */
    confPtr_t sampleExpansionDirection(Node * directionNode);

    /**
     * RRT Extend.
     * @param fromNode: node from which the expansion is performed
     * @param directionConf: configuration toward which the expansion is going
     * @param directionNode: node toward which the expansion is going (when relevant)
     * @return the number of created nodes
     */
    virtual unsigned extend(Node * fromNode, confPtr_t directionConf, Node * directionNode);

    /**
     * RRT Connect.
     * @param fromNode: node from which the expansion is performed
     * @param directionConf: configuration toward which the expansion is going
     * @param directionNode: node toward which the expansion is going (when relevant)
     * @return the number of created nodes
     */
    virtual unsigned connect(Node * fromNode, confPtr_t directionConf, Node * directionNode);

    /**
     * RRT Expand process: call the appropriate expansion method.
     * @param fromNode: node from which the expansion is performed
     * @param directionConf: configuration toward which the expansion is going
     * @param directionNode: node toward which the expansion is going (when relevant)
     * @param method: Extend or Connect
     * @return the number of created nodes
     */
    unsigned expandProcess(Node * fromNode, confPtr_t directionConf, Node * directionNode,
            Env::expansionMethod method);

    /**
     * Attempt to link two given nodes (i.e. two components, when relevant).
     * @return TRUE if the junction is successful.
     */
    virtual bool attemptJunction(Node * fromNode, Node * toNode);

    /**
     * Shoots a direction (includes the biasing)
     *
     * @param Expanding component
     * @param Goal Component
     * @param Sampling passive mode
     * @param Direction node
     */
    virtual std::tr1::shared_ptr<Configuration> getExpansionDirection(
            Node* expandComp, Node* goalComp, bool samplePassive,Node*& directionNode);

    /**
     * Gets the nearest node in the graph
     *
     * @param compNode Expanding component
     * @param direction Direction of expansion
     * @param Sampling passive mode
     */
    virtual Node* getExpansionNode(Node* compNode, confPtr_t direction, int distance);

private:
    //! goal node
    Node * goalNode;
};

#endif // RRTEXPANSION_HPP
