/*
 * Threshold-RRT.cpp
 *
 *  Created on: Feb 05, 2013
 *      Author: ddevaurs
 */

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "planner/Diffusion/Variants/Threshold-RRT.hpp"

using namespace std;


/**
 * ****************************************************
 *      Threshold Expansion
 * ****************************************************
 */

/**
 * Constructor.
 */
ThresholdExpansion::ThresholdExpansion(Graph * graph, Node * goal) : RRTExpansion(graph, goal)
{
	// if the threshold has been defined by the user, use that value
    if (ENV.getDouble(Env::costThreshold) > 0.)
    {
    	threshold = ENV.getDouble(Env::costThreshold);
    }
    // otherwise, use the minimal cost of nodes present in the graph
    else
    {
    	threshold = numeric_limits<double>::max();
    	vector<Node *> nodes = graph->getNodes();
    	for (vector<Node *>::iterator n = nodes.begin(); n != nodes.end(); ++n)
    		if ((*n)->cost() < threshold)
    			threshold = (*n)->cost();
    }

    // set the threshold increment
    increment = ENV.getDouble(Env::thresholdIncrement);
}


/**
 * Attempt to link two given nodes (i.e. two components, when relevant).
 * @return TRUE if the junction is successful.
 */
bool ThresholdExpansion::attemptJunction(Node * fromNode, Node * toNode)
{
    confPtr_t toConf = toNode->getConfiguration();

    // path along which the junction is attempted
    LocalPath path(fromNode->getConfiguration(), toConf);
    double pathLength = path.getParamMax();

    // N.B.: the connection is attempted only if the path length is less than a specified value
    // (which is some multiple of the extension step)
    if (pathLength <= ENV.getDouble(Env::minimalFinalExpansionGap) * this->step())
    {
        // check whether the path is valid (i.e. collision-free)
        if (!path.isValid())
        	return false;

        // interpolation factor and intermediate configuration
        double k = (pathLength == 0.) ? 1. : min(1., this->step() / pathLength);
        double prop(k);
        confPtr_t currentConf = (prop == 1.) ? toConf : path.configAtParam(prop * pathLength);

        // cost condition: is the cost of all intermediate configurations along the path below the threshold?
        while ((prop <= 1.) && (currentConf->cost() <= threshold))
        {
            if (prop < 1.)
            {
                prop = min(1., prop + k);
                currentConf = (prop == 1.) ? toConf : path.configAtParam(prop * pathLength);
            }
            else
                prop = 2.;
        }

        // if the cost condition is satisfied, validate the junction
        if (prop > 1.)
        	return true;
    }

    return false;
}


/**
 * RRT_obst Extend.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned ThresholdExpansion::extend(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{
	confPtr_t fromConf = fromNode->getConfiguration();
	ConnectedComponent * fromComp = fromNode->getConnectedComponent();

	// increase the cost threshold
    threshold += increment;

    // path along which the extension is performed
    LocalPath directionPath(fromConf, directionConf);
    double k = this->computeInterpolationFactor(directionPath.getParamMax());

    // Refinement control:
    // (k == 1. && !directionNode) means that we are potentially about to add a new refinement node;
    // in that case, if the number of refinement nodes already present in the component is too high
    // (i.e. more than 10% of the total number of nodes), the expansion attempt is canceled
    if (ENV.getBool(Env::refinementControl) && (k == 1. && !directionNode) &&
            (fromComp->getNumberOfRefinementNodes() > 0.1 * fromComp->getNumberOfNodes()))
        return 0;

    // construct an 'extension path', by linear interpolation along the 'direction path'
    LocalPath extensionPath(fromConf, this->interpolate(directionConf, &directionPath, k));

    // if the extension path is valid (i.e. without collision) and if the cost
    // of the new configuration is below the threshold, add a new node to the graph
    unsigned nbCreatedNodes(0);
    if (extensionPath.isValid() && extensionPath.getEnd()->cost() <= threshold)
    	this->addNode(fromNode, extensionPath, k, directionNode, nbCreatedNodes);

	// Refinement control:
	// if a new refinement node has been built, update the number of refinement nodes of the component
	if (ENV.getBool(Env::refinementControl) && (k == 1.) && (nbCreatedNodes > 0))
		fromComp->updateNumberOfRefinementNodes(nbCreatedNodes);

	return nbCreatedNodes;
}


/**
 * RRT_obst Connect.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned ThresholdExpansion::connect(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{
    confPtr_t fromConf = fromNode->getConfiguration();
    ConnectedComponent * fromComp = fromNode->getConnectedComponent();

	// increase the cost threshold
    threshold += increment;

    // path along which the connection is performed
    LocalPath directionPath(fromConf, directionConf);
    double k = this->computeInterpolationFactor(directionPath.getParamMax());

    // Refinement control:
    // (k == 1. && !directionNode) means that we are potentially about to add a new refinement node;
    // in that case, if the number of refinement nodes already present in the component is too high
    // (i.e. more than 10% of the total number of nodes), the expansion attempt is canceled
    if (ENV.getBool(Env::refinementControl) && (k == 1. && !directionNode) &&
            (fromComp->getNumberOfRefinementNodes() > 0.1 * fromComp->getNumberOfNodes()))
        return 0;

    bool success(true);
    confPtr_t currentConf = fromConf;
    confPtr_t newConf;
    double prop(k);
    double oldProp(0.);

    // attempt the connection step by step
    while (success && prop <= 1.)
    {
    	// check whether the current extension path is valid (i.e. without collision)
    	newConf = this->interpolate(directionConf, &directionPath, prop);
        LocalPath extensionPath(currentConf, newConf);
        if (!extensionPath.isValid())
            success = false;

        // check whether the cost of the new configuration is below the threshold
        if (success && (newConf->cost() <= threshold))
        {
            currentConf = newConf;
            oldProp = prop;
            if (prop < 1.)
                prop = min(1., prop + k);
            else
                prop = 2.;
        }
        else
            success = false;
    }

    // if "reached" or "advanced", add a node to the graph
    unsigned nbCreatedNodes(0);
    if (success || prop > k)
    {
        LocalPath connectPath(fromConf, currentConf);
        this->addNode(fromNode, connectPath, oldProp, directionNode, nbCreatedNodes);
    }

	// Refinement control:
	// if a refinement node has been built, update the number of refinement nodes of the component
	if (ENV.getBool(Env::refinementControl) && (k == 1.) && (nbCreatedNodes > 0))
		fromComp->updateNumberOfRefinementNodes(nbCreatedNodes);

    return nbCreatedNodes;
}


/**
 * ****************************************************
 *      Threshold RRT
 * ****************************************************
 */

/**
 * Initialize the RRT_obst.
 * @return the number of nodes added to the graph
 */
unsigned ThresholdRRT::init()
{
    // create the RRT_obst expansion
    expansion = new ThresholdExpansion(_Graph, _Goal);

    // initialize the Tree Planner
    return TreePlanner::init();
}
